@@ -20,7 +20,7 @@ else :
 
 import scipy
 from extrack.tracking import extract_params, get_all_Bs, get_Ts_from_Bs, first_log_integrale_dif, log_integrale_dif
-def P_segment_len(Cs, LocErr, ds, Fs, TrMat, pBL=0.1, isBL = 1, cell_dims = [0.5], nb_substeps=1, max_nb_states = 1000) :
+def P_segment_len(Cs, LocErr, ds, Fs, TrMat, min_l = 5, pBL=0.1, isBL = 1, cell_dims = [0.5], nb_substeps=1, max_nb_states = 1000) :
     '''
     compute the product of the integrals over Ri as previousily described
     work in log space to avoid overflow and underflow
@@ -45,12 +45,10 @@ def P_segment_len(Cs, LocErr, ds, Fs, TrMat, pBL=0.1, isBL = 1, cell_dims = [0.5
     Cs = cp.array(Cs)
     nb_states = TrMat.shape[0]
     Cs = Cs[:,:,::-1]
-    min_l = 5
-    
+        
     cell_dims = np.array(cell_dims)
     cell_dims = cell_dims[cell_dims!=None]
     
-    
     cur_Bs = get_all_Bs(nb_substeps + 1, nb_states)[None]
     
     TrMat = cp.array(TrMat.T)
@@ -113,7 +111,7 @@ def P_segment_len(Cs, LocErr, ds, Fs, TrMat, pBL=0.1, isBL = 1, cell_dims = [0.5
         p_stay = np.ones(sub_ds.shape[-1])
         for cell_len in cell_dims:
             xs = np.linspace(0+cell_len/2000,cell_len-cell_len/2000,1000)
-            cur_p_stay = ((np.mean(scipy.stats.norm.cdf((cell_len-xs[:,None])/(sub_ds+1e-200)) - scipy.stats.norm.cdf(-xs[:,None]/(sub_ds+1e-200)),0))*2)/2 # proba to stay in the FOV for each of the possible cur Bs
+            cur_p_stay = ((np.mean(scipy.stats.norm.cdf((cell_len-xs[:,None])/(sub_ds+1e-200)) - scipy.stats.norm.cdf(-xs[:,None]/(sub_ds+1e-200)),0))) # proba to stay in the FOV for each of the possible cur Bs
             p_stay = p_stay*cur_p_stay
         Lp_stay = np.log(p_stay * (1-pBL)) # proba for the track to survive = both stay in the FOV and not bleach
             
@@ -275,13 +273,14 @@ def len_hist(all_tracks,params, dt, cell_dims=[0.5,None,None], nb_states=2, nb_s
     here sum the logs(likelihood) to avoid too big numbers
     '''
     LocErr, ds, Fs, TrMat, pBL = extract_params(params, dt, nb_states, nb_substeps)
-
+    min_l = np.min((np.array(list(all_tracks.keys()))).astype(int))
+    
     if type(all_tracks) == type({}):
         new_all_tracks = []
         for l in all_tracks:
             new_all_tracks.append(all_tracks[l])
         all_tracks = new_all_tracks
-    
+       
     #seg_len_hists = np.zeros((all_tracks[-1].shape[1]+1,nb_states))
     seg_len_hists = np.zeros((all_tracks[-1].shape[1],nb_states))
     for k in range(len(all_tracks)):
@@ -296,7 +295,7 @@ def len_hist(all_tracks,params, dt, cell_dims=[0.5,None,None], nb_states=2, nb_s
             nb_max = 50
             for n in range(int(np.ceil(len(Css)/nb_max))):
                 Csss = Css[n*nb_max:(n+1)*nb_max]
-                LP, cur_Bs, seg_len_hist  = P_segment_len(Csss, LocErr, ds, Fs, TrMat, pBL=pBL, isBL = isBL, cell_dims = cell_dims, nb_substeps=nb_substeps, max_nb_states = max_nb_states)
+                LP, cur_Bs, seg_len_hist  = P_segment_len(Csss, LocErr, ds, Fs, TrMat, min_l = min_l, pBL=pBL, isBL = isBL, cell_dims = cell_dims, nb_substeps=nb_substeps, max_nb_states = max_nb_states)
                 isBL = 0
                 seg_len_hists[:seg_len_hist.shape[0]] = seg_len_hists[:seg_len_hist.shape[0]] + seg_len_hist
     print('')
