@@ -1,3 +1,9 @@
+# -*- coding: utf-8 -*-
+"""
+Created on Fri Apr  1 11:28:38 2022
+
+@author: Franc
+"""
 
 import numpy as np
 
@@ -15,6 +21,9 @@ import itertools
 import scipy
 from lmfit import minimize, Parameters
 
+import multiprocessing
+from itertools import product
+
 '''
 Maximum likelihood to determine transition rates :
 We compute the probability of observing the tracks knowing the parameters :
@@ -30,37 +39,37 @@ to compute the probability of having the tracks.
 '''
 
 def ds_froms_states(ds, cur_states):
-    cur_ds = ds[cur_states]
-    cur_ds = (cur_ds[:,:,1:]**2 + cur_ds[:,:,:-1]**2)**0.5 / 2**0.5 # assuming a transition at the middle of the substeps
+    cur_d2s = ds[cur_states]**2
+    cur_d2s = (cur_d2s[:,:,1:] + cur_d2s[:,:,:-1]) / 2 # assuming a transition at the middle of the substeps
     # we can average the variances of displacements per step to get the actual std of displacements per step
-    cur_ds = cp.mean(cur_ds**2, axis = 2)**0.5
-    cur_ds = cur_ds[:,:,None]
-    cur_ds = cp.array(cur_ds)
-    return cur_ds
-    
-def log_integrale_dif(Ci, l, cur_ds, Km, Ks):
+    cur_d2s = cp.mean(cur_d2s, axis = 2)
+    cur_d2s = cur_d2s[:,:,None]
+    cur_d2s = cp.array(cur_d2s)
+    return cur_d2s
+
+def log_integrale_dif(Ci, l2, cur_d2s, m_arr, s2_arr):
     '''
     integral of the 3 exponetional terms (localization error, diffusion, previous term)
-    the integral over r1 of f_l(r1-c1)f_d(r1-r0)f_Ks(r1-Km) equals :
-    np.exp(-((l**2+Ks**2)*r0**2+(-2*Km*l**2-2*Ks**2*c1)*r0+Km**2*l**2+(Km**2-2*c1*Km+c1**2)*d**2+Ks**2*c1**2)/((2*d**2+2*Ks**2)*l**2+2*Ks**2*d**2))/(2*np.pi*Ks*d*l*np.sqrt((d**2+Ks**2)*l**2+Ks**2*d**2))
-    which can be turned into the form Constant*fKs(r0 - newKm) where fKs is a normal law of std newKs
+    the integral over r1 of f_l(r1-c1)f_d(r1-r0)f_Ks(r1-m_arr) equals :
+    np.exp(-((l**2+Ks**2)*r0**2+(-2*m_arr*l**2-2*Ks**2*c1)*r0+m_arr**2*l**2+(m_arr**2-2*c1*m_arr+c1**2)*d**2+Ks**2*c1**2)/((2*d**2+2*Ks**2)*l**2+2*Ks**2*d**2))/(2*np.pi*Ks*d*l*np.sqrt((d**2+Ks**2)*l**2+Ks**2*d**2))
+    which can be turned into the form Constant*fKs(r0 - newm_arr) where fKs is a normal law of std newKs
     the idea is to create a function of integral of integral of integral etc
     dim 0 : tracks
     dim 1 : possible sequences of states
     dim 2 : x,y (z)
     '''
-    newKm = (Km*l**2 + Ci*Ks**2)/(l**2+Ks**2)
-    newKs = ((cur_ds**2*l**2 + cur_ds**2*Ks**2 + l**2*Ks**2)/(l**2 + Ks**2))**0.5
-    logConstant = Km.shape[2]*cp.log(1/((l**2 + Ks[:,:,0]**2)**0.5*np.sqrt(2*np.pi))) + cp.sum((newKm**2/(2*newKs**2) - (Km**2*l**2 + Ks**2*Ci**2 + (Km-Ci)**2*cur_ds**2)/(2*newKs**2*(l**2 + Ks**2))), axis = 2)
-    return newKm, newKs, logConstant
+    new_m = (m_arr*l2 + Ci*s2_arr)/(l2+s2_arr)
+    new_s2 = ((cur_d2s*l2 + cur_d2s*s2_arr + l2*s2_arr)/(l2 + s2_arr))
+    new_K = m_arr.shape[2] * -0.5*cp.log(2*np.pi*(l2 + s2_arr[:,:,0])) - cp.sum((Ci-m_arr)**2/(2*(l2+s2_arr)),axis = 2)
+    return new_m, new_s2, new_K
 
-def first_log_integrale_dif(Ci, LocErr, cur_ds):
+def first_log_integrale_dif(Ci, l2, cur_d2s):
     '''
     convolution of 2 normal laws = normal law (mean = sum of means and variance = sum of variances)
     '''
-    Ks = (LocErr**2+cur_ds**2)**0.5
-    Km = Ci
-    return Km, Ks
+    s2_arr = l2+cur_d2s
+    m_arr = Ci
+    return m_arr, s2_arr
 
 def P_Cs_inter_bound_stats(Cs, LocErr, ds, Fs, TrMat, pBL=0.1, isBL = 1, cell_dims = [0.5], nb_substeps=1, frame_len = 6, do_preds = 0, min_len = 3) :
     '''
@@ -71,13 +80,13 @@ def P_Cs_inter_bound_stats(Cs, LocErr, ds, Fs, TrMat, pBL=0.1, isBL = 1, cell_di
     dim 3 : x, y position
     
     we process by steps, at each step we account for 1 more localization, we compute
-    the canstant (LC), the mean (Km) and std (Ks) of of the normal distribution 
+    the canstant (LC), the mean (m_arr) and std (Ks) of of the normal distribution 
     resulting from integration.
     
     each step is made of substeps if nb_substeps > 1, and we increase the matrix
     of possible Bs : cur_Bs accordingly
     
-    to be able to process long tracks with good accuracy, for each track we fuse Km and Ks
+    to be able to process long tracks with good accuracy, for each track we fuse m_arr and Ks
     of sequences of states equal exept for the state 'frame_len' steps ago.
     '''
     nb_Tracks = Cs.shape[0]
@@ -87,6 +96,7 @@ def P_Cs_inter_bound_stats(Cs, LocErr, ds, Fs, TrMat, pBL=0.1, isBL = 1, cell_di
     Cs = cp.array(Cs)
     nb_states = TrMat.shape[0]
     Cs = Cs[:,:,::-1]
+    LocErr2 = LocErr**2
     
     if do_preds:
         preds = np.zeros((nb_Tracks, nb_locs, nb_states))-1
@@ -116,27 +126,29 @@ def P_Cs_inter_bound_stats(Cs, LocErr, ds, Fs, TrMat, pBL=0.1, isBL = 1, cell_di
     LP = LT + LF #+ compensate_leaving
     # current log proba of seeing the track
     LP = cp.repeat(LP, nb_Tracks, axis = 0)
-    cur_ds = ds[cur_states]
-    cur_ds = (cur_ds[:,:,1:]**2 + cur_ds[:,:,:-1]**2)**0.5 / 2**0.5 # assuming a transition at the middle of the substeps
+    cur_d2s = ds[cur_states]**2
+    cur_d2s = (cur_d2s[:,:,1:] + cur_d2s[:,:,:-1]) / 2 # assuming a transition at the middle of the substeps
     # we can average the variances of displacements per step to get the actual std of displacements per step
-    cur_ds = cp.mean(cur_ds**2, axis = 2)**0.5
-    cur_ds = cur_ds[:,:,None]
-    cur_ds = cp.array(cur_ds)
+    cur_d2s = cp.mean(cur_d2s, axis = 2)
+    cur_d2s = cur_d2s[:,:,None]
+    cur_d2s = cp.array(cur_d2s)
     
     sub_Bs = cur_Bs.copy()[:,:cur_Bs.shape[1]//nb_states,:nb_substeps] # list of possible current states we can meet to compute the proba of staying in the FOV
     sub_ds = cp.mean(ds[sub_Bs]**2, axis = 2)**0.5 # corresponding list of d
+    sub_ds = asnumpy(sub_ds)
     
     p_stay = np.ones(sub_ds.shape[-1])
     for cell_len in cell_dims:
         xs = np.linspace(0+cell_len/2000,cell_len-cell_len/2000,1000)
-        cur_p_stay = ((np.mean(scipy.stats.norm.cdf((cell_len-xs[:,None])/(sub_ds+1e-200)) - scipy.stats.norm.cdf(-xs[:,None]/(sub_ds+1e-200)),0))*2)/2 # proba to stay in the FOV for each of the possible cur Bs
+        cur_p_stay = ((cp.mean(scipy.stats.norm.cdf((cell_len-xs[:,None])/(sub_ds+1e-200)) - scipy.stats.norm.cdf(-xs[:,None]/(sub_ds+1e-200)),0))) # proba to stay in the FOV for each of the possible cur Bs
         p_stay = p_stay*cur_p_stay
-    Lp_stay = np.log(p_stay * (1-pBL)) # proba for the track to survive = both stay in the FOV and not bleach
+    p_stay = cp.array(p_stay)
+    Lp_stay = cp.log(p_stay * (1-pBL)) # proba for the track to survive = both stay in the FOV and not bleach
     
-    # inject the first position to get the associated Km and Ks :
-    Km, Ks = first_log_integrale_dif(Cs[:,:, nb_locs-current_step], LocErr, cur_ds)
+    # inject the first position to get the associated m_arr and Ks :
+    m_arr, s2_arr = first_log_integrale_dif(Cs[:,:, nb_locs-current_step], LocErr2, cur_d2s)
     current_step += 1
-    Km = cp.repeat(Km, cur_nb_Bs, axis = 1)
+    m_arr = cp.repeat(m_arr, cur_nb_Bs, axis = 1)
     removed_steps = 0
     #TrMat = np.array([[0.9,0.1],[0.2,0.8]])
     while current_step <= nb_locs-1:
@@ -146,20 +158,20 @@ def P_Cs_inter_bound_stats(Cs, LocErr, ds, Fs, TrMat, pBL=0.1, isBL = 1, cell_di
         
         cur_states = cur_Bs[:,:,0:nb_substeps+1].astype(int)
         # compute the vector of diffusion stds knowing the states at the current step
-        cur_ds = ds[cur_states]
-        cur_ds = (cur_ds[:,:,1:]**2 + cur_ds[:,:,:-1]**2)**0.5 / 2**0.5 # assuming a transition at the middle of the substeps
-        cur_ds = cp.mean(cur_ds**2, axis = 2)**0.5
-        cur_ds = cur_ds[:,:,None]
+        cur_d2s = ds[cur_states]**2
+        cur_d2s = (cur_d2s[:,:,1:] + cur_d2s[:,:,:-1]) / 2 # assuming a transition at the middle of the substeps
+        cur_d2s = cp.mean(cur_d2s, axis = 2)
+        cur_d2s = cur_d2s[:,:,None]
         LT = get_Ts_from_Bs(cur_states, TrMat)
 
         #np.arange(32)[None][:,TT]
         #np.arange(32)[None][:,1:][:,TT[:-1]]
         # repeat the previous matrix to account for the states variations due to the new position
-        Km = cp.repeat(Km, nb_states**nb_substeps , axis = 1)
-        Ks = cp.repeat(Ks, nb_states**nb_substeps, axis = 1)
+        m_arr = cp.repeat(m_arr, nb_states**nb_substeps , axis = 1)
+        s2_arr = cp.repeat(s2_arr, nb_states**nb_substeps, axis = 1)
         LP = cp.repeat(LP, nb_states**nb_substeps, axis = 1)
-        # inject the next position to get the associated Km, Ks and Constant describing the integral of 3 normal laws :
-        Km, Ks, LC = log_integrale_dif(Cs[:,:,nb_locs-current_step], LocErr, cur_ds, Km, Ks)
+        # inject the next position to get the associated m_arr, s2_arr and Constant describing the integral of 3 normal laws :
+        m_arr, s2_arr, LC = log_integrale_dif(Cs[:,:,nb_locs-current_step], LocErr2, cur_d2s, m_arr, s2_arr)
         #print('integral',time.time() - t0); t0 = time.time()
         if current_step >= min_len :
             LL = Lp_stay[np.argmax(np.all(cur_states[:,None,:,:-1] == sub_Bs[:,:,None],-1),1)] # pick the right proba of staying according to the current states
@@ -171,14 +183,14 @@ def P_Cs_inter_bound_stats(Cs, LocErr, ds, Fs, TrMat, pBL=0.1, isBL = 1, cell_di
         cur_nb_Bs = len(cur_Bs[0]) # current number of sequences of states
         
         ''''idea : the position and the state 6 steps ago should not impact too much the 
-        probability of the next position so the Km and Ks of tracks with the same 6 last 
+        probability of the next position so the m_arr and s2_arr of tracks with the same 6 last 
         states must be very similar, we can then fuse the parameters of the pairs of Bs
         which vary only for the last step (7) and sum their probas'''
         if current_step < nb_locs-1:
             while cur_nb_Bs >= nb_states**frame_len:
                 if do_preds :
-                    newKs = cp.array((Ks**2 + LocErr**2)**0.5)[:,:,0]
-                    log_integrated_term = -cp.log(2*np.pi*newKs**2) - cp.sum((Cs[:,:,nb_locs-current_step] - Km)**2,axis=2)/(2*newKs**2)
+                    new_s2_arr = cp.array((s2_arr + LocErr2))[:,:,0]
+                    log_integrated_term = -cp.log(2*np.pi*new_s2_arr) - cp.sum((Cs[:,:,nb_locs-current_step] - m_arr)**2,axis=2)/(2*new_s2_arr)
                     LF = 0 #cp.log(Fs[cur_Bs[:,:,0].astype(int)]) # Log proba of starting in a given state (fractions)
                     
                     test_LP = LP + log_integrated_term + LF
@@ -192,7 +204,7 @@ def P_Cs_inter_bound_stats(Cs, LocErr, ds, Fs, TrMat, pBL=0.1, isBL = 1, cell_di
                         B_is_state = cur_Bs[:,:,-1] == state
                         preds[:,nb_locs-current_step+frame_len-2, state] = asnumpy(np.sum(B_is_state*P,axis = 1)/np.sum(P,axis = 1))
                 cur_Bs = cur_Bs[:,:cur_nb_Bs//nb_states, :-1]
-                Km, Ks, LP = fuse_tracks(Km, Ks, LP, cur_nb_Bs, nb_states)
+                m_arr, s2_arr, LP = fuse_tracks(m_arr, s2_arr, LP, cur_nb_Bs, nb_states)
                 cur_nb_Bs = len(cur_Bs[0])
                 removed_steps += 1
         #print('frame',time.time() - t0)
@@ -211,18 +223,18 @@ def P_Cs_inter_bound_stats(Cs, LocErr, ds, Fs, TrMat, pBL=0.1, isBL = 1, cell_di
         LT = get_Ts_from_Bs(cur_states, TrMat)
         #cur_states = cur_states[:,:,0]
         # repeat the previous matrix to account for the states variations due to the new position
-        Km = cp.repeat(Km, nb_states**nb_substeps , axis = 1)
-        Ks = cp.repeat(Ks, nb_states**nb_substeps, axis = 1)
+        m_arr = cp.repeat(m_arr, nb_states**nb_substeps , axis = 1)
+        s2_arr = cp.repeat(s2_arr, nb_states**nb_substeps, axis = 1)
         LP = cp.repeat(LP, nb_states**nb_substeps, axis = 1)
         
         #LL = Lp_stay[np.argmax(np.all(cur_states[:,None] == sub_Bs[:,:,None],-1),1)] # pick the right proba of staying according to the current states
         #end_p_stay = p_stay[np.argmax(np.all(cur_states[:,None:,:-1] == sub_Bs[:,:,None],-1),1)]
         end_p_stay = p_stay[cur_states[:,None:,:-1]][:,:,0]
         end_p_stay.shape
-        LL = np.log(pBL + (1-end_p_stay) - pBL * (1-end_p_stay)) + LT
+        LL = cp.log(pBL + (1-end_p_stay) - pBL * (1-end_p_stay)) + LT
         
-    newKs = cp.array((Ks**2 + LocErr**2)**0.5)[:,:,0]
-    log_integrated_term = -cp.log(2*np.pi*newKs**2) - cp.sum((Cs[:,:,0] - Km)**2,axis=2)/(2*newKs**2)
+    new_s2_arr = cp.array((s2_arr + LocErr2))[:,:,0]
+    log_integrated_term = -cp.log(2*np.pi*new_s2_arr) - cp.sum((Cs[:,:,0] - m_arr)**2,axis=2)/(2*new_s2_arr)
     #LF = cp.log(Fs[cur_Bs[:,:,0].astype(int)]) # Log proba of starting in a given state (fractions)
     #LF = cp.log(0.5)
     # cp.mean(cp.log(Fs[cur_Bs[:,:,:].astype(int)]), 2) # Log proba of starting in a given state (fractions)
@@ -240,10 +252,10 @@ def P_Cs_inter_bound_stats(Cs, LocErr, ds, Fs, TrMat, pBL=0.1, isBL = 1, cell_di
         preds = preds[:,::-1]
     return LP, cur_Bs, preds
 
-def fuse_tracks(Km, Ks, LP, cur_nb_Bs, nb_states = 2):
+def fuse_tracks(m_arr, s2_arr, LP, cur_nb_Bs, nb_states = 2):
     '''
     The probabilities of the pairs of tracks must be added
-    I chose to define the updated Km and Ks as the weighted average (of the variance for Ks)
+    I chose to define the updated m_arr and s2_arr as the weighted average (of the variance for s2_arr)
     but other methods may be better
     As I must divid by a sum of exponentials which can be equal to zero because of underflow
     I correct the values in the exponetial to keep the maximal exp value at 0
@@ -251,16 +263,16 @@ def fuse_tracks(Km, Ks, LP, cur_nb_Bs, nb_states = 2):
     # cut the matrixes so the resulting matrices only vary for their last state
     I = cur_nb_Bs//nb_states
     LPk = []
-    Kmk = []
-    Ksk = []
+    m_arr_k = []
+    s2_arr_k = []
     for k in range(nb_states):
         LPk.append(LP[:, k*I:(k+1)*I])# LP of which the last state is k
-        Kmk.append(Km[:, k*I:(k+1)*I])# Km of which the last state is k
-        Ksk.append(Ks[:, k*I:(k+1)*I])# Ks of which the last state is k
+        m_arr_k.append(m_arr[:, k*I:(k+1)*I])# m_arr of which the last state is k
+        s2_arr_k.append(s2_arr[:, k*I:(k+1)*I])# s2_arr of which the last state is k
         
     LPk = cp.array(LPk)
-    Kmk = cp.array(Kmk)
-    Ksk = cp.array(Ksk)
+    m_arr_k = cp.array(m_arr_k)
+    s2_arr_k = cp.array(s2_arr_k)
     
     maxLP = cp.max(LPk, axis = 0, keepdims = True)
     Pk = cp.exp(LPk - maxLP)
@@ -270,15 +282,15 @@ def fuse_tracks(Km, Ks, LP, cur_nb_Bs, nb_states = 2):
     ak = Pk/SP
     
     # update the parameters, this step is tricky as an approximation of a gaussian mixture by a simple gaussian
-    Km = cp.sum(ak[:,:,:,None] * Kmk, axis=0)
-    Ks = cp.sum((ak[:,:,:,None] * Ksk**2), axis=0)**0.5
+    m_arr = cp.sum(ak[:,:,:,None] * m_arr_k, axis=0)
+    s2_arr = cp.sum((ak[:,:,:,None] * s2_arr_k), axis=0)
     del ak
     LP = maxLP + np.log(SP)
     LP = LP[0]
     # cur_Bs = cur_Bs[:,:I, :-1]
-    # np.mean(np.abs(Km0-Km1)) # to verify how far they are, I found a difference of 0.2nm for D = 0.1um2/s, LocErr=0.02um and 6 frames
-    # np.mean(np.abs(Ks0-Ks1))
-    return Km, Ks, LP
+    # np.mean(np.abs(m_arr0-m_arr1)) # to verify how far they are, I found a difference of 0.2nm for D = 0.1um2/s, LocErr=0.02um and 6 frames
+    # np.mean(np.abs(s2_arr0-s2_arr1))
+    return m_arr, s2_arr, LP
 
 def get_all_Bs(nb_Cs, nb_states):
     '''
@@ -353,13 +365,14 @@ def predict_Bs(all_tracks,
     max_len = int(l_list[-1])
     
     for l in l_list:
-        if int(l) == max_len:
-            isBL = 1
-        else:
-            isBL = 0
-        LP_Cs, trunkated_Bs, pred_Bs = P_Cs_inter_bound_stats(all_tracks[l], LocErr, ds, Fs, TrMat, pBL,isBL, cell_dims, nb_substeps = 1, frame_len = frame_len, do_preds = 1, min_len = min_len)
-        all_pred_Bs.append(pred_Bs)
-        
+        if len(all_tracks[l]) > 0:
+            if int(l) == max_len:
+                isBL = 1
+            else:
+                isBL = 0
+            LP_Cs, trunkated_Bs, pred_Bs = P_Cs_inter_bound_stats(all_tracks[l], LocErr, ds, Fs, TrMat, pBL,isBL, cell_dims, nb_substeps = 1, frame_len = frame_len, do_preds = 1, min_len = min_len)
+            all_pred_Bs.append(pred_Bs)
+    
     all_pred_Bs_dict = {}
     for pred_Bs in all_pred_Bs:
         all_pred_Bs_dict[str(pred_Bs.shape[1])] = pred_Bs
@@ -373,76 +386,48 @@ def extract_params(params, dt, nb_states, nb_substeps):
     '''
     LocErr = params['LocErr'].value
     
-    if nb_states == 2:
-        D0 = params['D0'].value
-        D1 = params['D1'].value
-        Ds = np.array([D0,D1])
-        F0 = params['F0'].value
-        Fs = np.array([F0, 1-F0])
-        pBL = params['pBL'].value
-        p01 = params['p01']
-        p10 =  params['p10']
+    param_names = np.sort(list(params.keys()))
 
-        # correct p10 and p01 which actually correspond to r*dt to get the 
-        # corresponding discrete probabilities of at least 1 transition during 
-        # a substep :
-        p01 = 1 - np.exp(-p01/nb_substeps)
-        p10 = 1 - np.exp(-p10/nb_substeps)
-        
-        TrMat = np.array([[1-p01, p01],[p10, 1- p10]])
+    Ds = []
+    Fs = []
+    for param in param_names:
+        if param.startswith('D') and len(param)<3:
+            Ds.append(params[param].value)
+        elif param.startswith('F'):
+            Fs.append(params[param].value)
+    Ds = np.array(Ds)
+    Fs = np.array(Fs)
+    TrMat = np.zeros((len(Ds),len(Ds)))
+    for param in params:
+        if param == 'pBL':
+            pBL = params[param].value
+        elif param.startswith('p'):
+            i = int(param[1])
+            j = int(param[2])
+            TrMat[i,j] = params[param].value
+    
+    TrMat = TrMat/nb_substeps
     
-    elif nb_states == 3:
-        D0 = params['D0'].value
-        D1 = params['D1'].value
-        D2 = params['D2'].value
-        Ds = np.array([D0,D1,D2])
-        p01 = params['p01'].value
-        p02 = params['p02'].value
-        p10 = params['p10'].value
-        p12 = params['p12'].value
-        p20 = params['p20'].value
-        p21 = params['p21'].value
-        
-        F0 = params['F0'].value
-        F1 = params['F1'].value
-        F2 = params['F2'].value
-        pBL = params['pBL'].value
-        
-        Fs = np.array([F0, F1, F2])
-        p01 = 1 - np.exp(-p01/nb_substeps)
-        p02 = 1 - np.exp(-p02/nb_substeps)
-        p10 = 1 - np.exp(-p10/nb_substeps)
-        p12 = 1 - np.exp(-p12/nb_substeps)
-        p20 = 1 - np.exp(-p20/nb_substeps)
-        p21 = 1 - np.exp(-p21/nb_substeps)
 
-        TrMat = np.array([[1-p01-p02, p01, p02],
-                          [p10, 1-p10-p12, p12],
-                          [p20, p21, 1-p20-p21]])
-    else :
-        Ds = []
-        Fs = []
-        for param in params:
-            if param.startswith('D') and len(param)<3:
-                Ds.append(params[param].value)
-            elif param.startswith('F'):
-                Fs.append(params[param].value)
-        Ds = np.array(Ds)
-        Fs = np.array(Fs)
-        TrMat = np.zeros((len(Ds),len(Ds)))
-        for param in params:
-            if param == 'pBL':
-                pBL = params[param].value
-            elif param.startswith('p'):
-                i = int(param[1])
-                j = int(param[2])
-                TrMat[i,j] = params[param].value
-        TrMat = 1 - np.exp(-TrMat/nb_substeps)
-        TrMat[np.arange(len(Ds)), np.arange(len(Ds))] = 1-np.sum(TrMat,1)
+    TrMat = 1 - np.exp(-TrMat)
+    TrMat[np.arange(len(Ds)), np.arange(len(Ds))] = 1-np.sum(TrMat,1)
+    
+    #print(TrMat)
     ds = np.sqrt(2*Ds*dt)
     return LocErr, ds, Fs, TrMat, pBL
 
-def cum_Proba_Cs(params, all_tracks, dt, cell_dims, nb_states, nb_substeps, frame_len, verbose = 1):
+def pool_star_proba(args):
+    return Proba_Cs(*args)
+
+def run_pool(pool_star_proba, args_prod, workers):
+    with multiprocessing.Pool(workers) as pool:
+        LP = pool.map(pool_star_proba, args_prod)
+    return LP
+
+def f(x):
+    return x*x
+
+def cum_Proba_Cs(params, all_tracks, dt, cell_dims, nb_states, nb_substeps, frame_len, verbose = 1, workers = 1):
     '''
     each probability can be multiplied to get a likelihood of the model knowing
     the parameters LocErr, D0 the diff coefficient of state 0 and F0 fraction of
@@ -452,19 +437,38 @@ def cum_Proba_Cs(params, all_tracks, dt, cell_dims, nb_states, nb_substeps, fram
     '''
     LocErr, ds, Fs, TrMat, pBL = extract_params(params, dt, nb_states, nb_substeps)
     min_len = all_tracks[0].shape[1]
-    if np.all(TrMat>0):
+    
+    if np.all(TrMat>0) and np.all(Fs>0):
         Cum_P = 0
+        Csss = []
+        isBLs = []
         for k in range(len(all_tracks)):
             if k == len(all_tracks)-1:
                 isBL = 0 # last position correspond to tracks which didn't disapear within maximum track length
             else:
                 isBL = 1
             Css = all_tracks[k]
-            nb_max = 700
+            nb_max = 50
             for n in range(int(np.ceil(len(Css)/nb_max))):
-                Csss = Css[n*nb_max:(n+1)*nb_max]
-                LP = Proba_Cs(Csss, LocErr, ds, Fs, TrMat,pBL, isBL,cell_dims, nb_substeps, frame_len, min_len)
-                Cum_P += cp.sum(LP)
+                Csss.append(Css[n*nb_max:(n+1)*nb_max])
+                if k == len(all_tracks)-1:
+                    isBLs.append(0) # last position correspond to tracks which didn't disapear within maximum track length
+                else:
+                    isBLs.append(1)
+        Csss.reverse()
+        args_prod = np.array(list(product(Csss, [LocErr], [ds], [Fs], [TrMat],[pBL], [0],[cell_dims], [nb_substeps], [frame_len], [min_len])))
+        args_prod[:, 6] = isBLs
+        
+        if workers >= 2:
+            with multiprocessing.Pool(workers) as pool:
+                LP = pool.map(pool_star_proba, args_prod)
+
+        else:
+            LP = []
+            for args in args_prod:
+                LP.append(pool_star_proba(args))
+            
+        Cum_P += cp.sum(cp.concatenate(LP))
         Cum_P = asnumpy(Cum_P)
         
         if verbose == 1:
@@ -475,7 +479,7 @@ def cum_Proba_Cs(params, all_tracks, dt, cell_dims, nb_states, nb_substeps, fram
         out = - Cum_P # normalize by the number of tracks and number of displacements
     else:
         out = 1e100
-        print('wrong set of inputs')
+        print('x',end='')
     if out == np.nan:
         out = 1e100
         print('inputs give nans')
@@ -496,6 +500,7 @@ def get_params(nb_states = 2,
                             {'name' : 'D1', 'expr' : 'D0 + D1_minus_D0'},
                             {'name' : 'LocErr', 'value' : estimated_vals['LocErr'], 'min' :  min_values['LocErr'],'max' :  max_values['LocErr'], 'vary' : vary_params['LocErr']},
                             {'name' : 'F0', 'value' : estimated_vals['F0'], 'min' :  min_values['F0'], 'max' :  max_values['F0'], 'vary' :  vary_params['F0']},
+                            {'name' : 'F1', 'expr' : '1 - F0'},
                             {'name' : 'p01', 'value' : estimated_vals['p01'], 'min' :  min_values['p01'], 'max' :  max_values['p01'], 'vary' :  vary_params['p01']},
                             {'name' : 'p10', 'expr' : 'p01/(1/F0-1)'},
                             {'name' : 'pBL', 'value' : estimated_vals['pBL'], 'min' :  min_values['pBL'], 'max' :  max_values['pBL'], 'vary' : vary_params['pBL']}]
@@ -505,6 +510,7 @@ def get_params(nb_states = 2,
                             {'name' : 'D1', 'expr' : 'D0 + D1_minus_D0' },
                             {'name' : 'LocErr', 'value' : estimated_vals['LocErr'], 'min' :  min_values['LocErr'],'max' :  max_values['LocErr'], 'vary' : vary_params['LocErr']},
                             {'name' : 'F0', 'value' : estimated_vals['F0'], 'min' :  min_values['F0'], 'max' :  max_values['F0'], 'vary' :  vary_params['F0']},
+                            {'name' : 'F1', 'expr' : '1 - F0'},
                             {'name' : 'p01', 'value' : estimated_vals['p01'], 'min' :  min_values['p01'], 'max' :  max_values['p01'], 'vary' :  vary_params['p01']},
                             {'name' : 'p10', 'value' : estimated_vals['p10'], 'min' :  min_values['p10'], 'max' :  max_values['p10'], 'vary' : vary_params['p10']},
                             {'name' : 'pBL', 'value' : estimated_vals['pBL'], 'min' :  min_values['pBL'], 'max' :  max_values['pBL'], 'vary' : vary_params['pBL']}]
@@ -567,15 +573,17 @@ def get_params(nb_states = 2,
             param_kwargs.append({'name' : D, 'expr' : expr})
             last_D = D
             sum_Ds += estimated_vals[D]
+        
         param_kwargs.append({'name' : 'F0', 'value' : estimated_vals['F0'], 'min' : min_values['F0'], 'max' : 0.3, 'brute_step' : 0.04, 'vary' : vary_params['F0']})
         frac = 1-estimated_vals['F0']
-        expr = '1-F0'
-        for F in Fs[1:]:
-            param_kwargs.append({'name' : F + '_fract', 'value' : estimated_vals[F]/frac, 'min' : 0, 'max' : 1 , 'vary' : vary_params[F]})
-            param_kwargs.append({'name' : F, 'expr' : F + '_fract*('+expr+')'})
+        expr = '1-F0'        
+        
+        for F in Fs[1:len(Ds)-1]:
+            param_kwargs.append({'name' : F , 'value' : estimated_vals[F], 'min' : 0.001, 'max' : 0.99 , 'vary' : vary_params[F]})
+            frac = frac - 1
             expr = expr + '-' + F
-            frac = frac - estimated_vals[F]
-        param_kwargs.append({'name' : 'F'+str(len(Fs)), 'expr' : expr})
+        param_kwargs.append({'name' : 'F'+str(len(Ds)-1), 'expr' : expr})
+        
         for param in list(vary_params.keys()):
             if param.startswith('p'):
                 param_kwargs.append({'name' : param, 'value' : estimated_vals[param], 'min' : min_values[param], 'max' : max_values[param] , 'vary' : vary_params[param]})
@@ -590,6 +598,7 @@ def get_2DSPT_params(all_tracks,
                      nb_states = 2,
                      frame_len = 8,
                      verbose = 1,
+                     workers = 1,
                      method = 'powell',
                      steady_state = False,
                      cell_dims = [1], # list of dimensions limit for the field of view (FOV) of the cell in um, a membrane protein in a typical e-coli cell in tirf would have a cell_dims = [0.5,3], in case of cytosolic protein one should imput the depth of the FOV e.g. [0.3] for tirf or [0.8] for hilo
@@ -635,22 +644,22 @@ def get_2DSPT_params(all_tracks,
         if not (len(min_values) == 7 and len(max_values) == 7 and len(estimated_vals) == 7 and len(vary_params) == 7):
             raise ValueError('estimated_vals, min_values, max_values and vary_params should all containing 7 parameters')
     elif nb_states == 3:
-        if len(vary_params) == 7:
+        if len(vary_params) != 13:
             vary_params = {'LocErr' : True, 'D0' : True, 'D1' :  True, 'D2' : True, 'F0' : True, 'F1' : True, 'p01' : True, 'p02' : True, 'p10' : True,'p12' :  True,'p20' :  True, 'p21' : True, 'pBL' : True},
-        if len(estimated_vals) == 7:
+        if len(estimated_vals) != 13:
             estimated_vals = {'LocErr' : 0.023, 'D0' : 1e-20, 'D1' : 0.02, 'D2' :  0.1, 'F0' : 0.33,  'F1' : 0.33, 'p01' : 0.1, 'p02' : 0.1, 'p10' :0.1, 'p12' : 0.1, 'p20' :0.1, 'p21' :0.1, 'pBL' : 0.1},
-        if len(min_values) == 7:
+        if len(min_values) != 13:
             min_values = {'LocErr' : 0.007, 'D0' : 1e-20, 'D1' : 0.0000001, 'D2' :  0.000001, 'F0' : 0.001,  'F1' : 0.001, 'p01' : 0.001, 'p02' : 0.001, 'p10' :0.001, 'p12' : 0.001, 'p20' :0.001, 'p21' :0.001, 'pBL' : 0.001},
-        if len(max_values) == 7:
-            max_values = {'LocErr' : 0.023, 'D0' : 1, 'D1' : 1, 'D2' :  10, 'D3' : 100, 'F0' : 0.999,  'F1' : 0.999, 'F2' : 0.999, 'p01' : 1, 'p02' : 1, 'p03' : 1, 'p10' :1, 'p12' : 1, 'p13' : 1, 'p20' : 1, 'p21' : 1, 'p23' : 1, 'p30' : 1, 'p31' : 1, 'p32' : 1, 'pBL' : 0.99}
+        if len(max_values) != 13:
+            max_values = {'LocErr' : 0.023, 'D0' : 1, 'D1' : 1, 'D2' :  10, 'F0' : 0.999,  'F1' : 0.999, 'p01' : 1, 'p02' : 1, 'p10' :1, 'p12' : 1, 'p20' : 1, 'p21' : 1, 'pBL' : 0.99}
     elif nb_states == 4:
-        if len(vary_params) == 7:
+        if len(vary_params) != 21:
             vary_params = {'LocErr' : True, 'D0' : True, 'D1' : True, 'D2' :  True, 'D3' : True, 'F0' : True,  'F1' : True, 'F2' : True, 'p01' : True, 'p02' : True, 'p03' : True, 'p10' : True, 'p12' : True, 'p13' : True, 'p20' :True, 'p21' :True, 'p23' : True, 'p30' :True, 'p31' :True, 'p32' : True, 'pBL' : True}
-        if len(estimated_vals) == 7:
+        if len(estimated_vals) != 21:
             estimated_vals = {'LocErr' : 0.023, 'D0' : 1e-20, 'D1' : 0.02, 'D2' :  0.1, 'D3' : 0.5, 'F0' : 0.1,  'F1' : 0.2, 'F2' : 0.3, 'p01' : 0.1, 'p02' : 0.1, 'p03' : 0.1, 'p10' :0.1, 'p12' : 0.1, 'p13' : 0.1, 'p20' :0.1, 'p21' :0.1, 'p23' : 0.1, 'p30' :0.1, 'p31' :0.1, 'p32' : 0.1, 'pBL' : 0.1}
-        if len(min_values) == 7:
+        if len(min_values) != 21:
             min_values = {'LocErr' : 0.005, 'D0' : 0, 'D1' : 0, 'D2' :  0.001, 'D3' : 0.001, 'F0' : 0.001,  'F1' : 0.001, 'F2' : 0.001, 'p01' : 0.001, 'p02' : 0.001, 'p03' : 0.001, 'p10' :0.001, 'p12' : 0.001, 'p13' : 0.001, 'p20' :0.001, 'p21' :0.001, 'p23' : 0.001, 'p30' :0.001, 'p31' :0.001, 'p32' : 0.001, 'pBL' : 0.001}
-        if len(max_values) == 7:
+        if len(max_values) != 21:
             max_values = {'LocErr' : 0.023, 'D0' : 1, 'D1' : 1, 'D2' :  10, 'D3' : 100, 'F0' : 0.999,  'F1' : 0.999, 'F2' : 0.999, 'p01' : 1, 'p02' : 1, 'p03' : 1, 'p10' :1, 'p12' : 1, 'p13' : 1, 'p20' : 1, 'p21' : 1, 'p23' : 1, 'p30' : 1, 'p31' : 1, 'p32' : 1, 'pBL' : 0.99}
     else:
         if len(vary_params) == 7 or len(estimated_vals) == 7 or len(min_values) == 7 or len(max_values) == 7:
@@ -666,13 +675,10 @@ def get_2DSPT_params(all_tracks,
     l_list = np.sort(np.array(list(all_tracks.keys())).astype(int)).astype(str)
     sorted_tracks = []
     for l in l_list:
-        sorted_tracks.append(all_tracks[l])
+        if len(all_tracks[l]) > 0 :
+            sorted_tracks.append(all_tracks[l])
     all_tracks = sorted_tracks
     
-    fit = minimize(cum_Proba_Cs, params, args=(all_tracks, dt, cell_dims, nb_states, nb_substeps, frame_len, verbose), method = method, nan_policy = 'propagate')
+    fit = minimize(cum_Proba_Cs, params, args=(all_tracks, dt, cell_dims, nb_states, nb_substeps, frame_len, verbose, workers), method = method, nan_policy = 'propagate')
     
     return fit
-
-
-
-
